/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package it.uniroma1.dis.mfis.redia.verifint.view.graphics.engine;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.util.ArrayList;
import java.util.List;

/**
 * Classe che rappresenta e gestisce graficamente un multi arco. 
 * Un multi arco e' un collegamento attivo tra un minimo di due figure ad un 
 * massimo ideale di infinite figure.
 * @author Fuego
 */
public class Edge implements Drawable, Comparable {     
    private List<LinkPoint> startLinkPoints;
    private LinkPoint endingLinkPoint;
    private int stile = STYLE_ISA;
    // Stili presi da convenzione UML
    public static final int STYLE_ISA = 0;
    public static final int STYLE_IMPLEMENTS = 1;
    public static final int STYLE_SIMPLEASSOCIATION = 2;
    public static final int STYLE_HAS = 3;
    public static final int STYLE_USES = 4;
    public static final int STYLE_AGGREGATION = 5;
    public static final int STYLE_COMPOSITION = 6;
    private static final int LINKPOINT_RADIUS = 15;
    private static final String DEFAULT_LABEL = "";
    // Etichetta Nome dell'associazione
    private Label labelName = new Label(DEFAULT_LABEL);

    /**
     * Crea una nuova istanza di Arco a partire da sola una figura iniziale
     * ad una sola figura finale
     * @param start Figura da dove parte l'arco
     * @param end Figura dove termina l'arco
     */
    public Edge(Figure start, Figure end) {
        startLinkPoints = new ArrayList();
        this.startLinkPoints.add(new LinkPoint(start, null, DEFAULT_LABEL));
        this.endingLinkPoint = new LinkPoint(end, null, DEFAULT_LABEL);        
        labelName.setOpaque(true);
        calculateLinkPoints();
    }

    /**
     * Crea una nuova istanza di Arco a partire da una serie di figure iniziali
     * ad una sola figura finale. Rappresenta un arco multiplo.
     * @param start Lista di figure da cui parte l'arco
     * @param end Figura dove termina l'arco
     */
    public Edge(List<Figure> start, Figure end) {
        startLinkPoints = new ArrayList();
        for (Figure f : start) {
            startLinkPoints.add(new LinkPoint(f, null, DEFAULT_LABEL));           
        }                
        this.endingLinkPoint = new LinkPoint(end, null, DEFAULT_LABEL);    
        labelName.setOpaque(true);
        calculateLinkPoints();
    }

    /**
     * Posiziona i punti di aggancio e le etichette nel modo ottimale in base
     * alle nuove posizioni attuali delle figure associate.
     */
    public void optimize() {
        calculateLinkPoints();        
    }
    /**
     * Calcola automaticamente i punti di aggancio ottimali dell'arco.
     * I punti di aggancio sono i punti appartenenti al bordo della figura che
     * si vuole collegare
     */
    private void calculateLinkPoints() {  
        /* Calcolo la media dei baricentri delle figure, in questo modo
         * posso trovare il baricentro (cg) tra i tre punti, dove gli archi multipli
         * si incontreranno
         */
        int sumx = endingLinkPoint.getFigure().getCG().x;
        int sumy = endingLinkPoint.getFigure().getCG().y;
        Point cg = new Point();
        for (LinkPoint lp : startLinkPoints) {
            sumx += lp.getFigure().getCG().x;
            sumy += lp.getFigure().getCG().y;            
        }
        cg.x = sumx / (startLinkPoints.size() + 1);
        cg.y = sumy / (startLinkPoints.size() + 1);

        /* Adesso trovo i punti di aggancio, cioe' quei punti che fanno parte
         * del bordo delle figure da collegare. Basta trovare il punto più vicino
         * tra il baricentro complessivo e la figura da agganciare!
         */
        for (LinkPoint lp : startLinkPoints) {
            lp.hookPoint(cg);            
        }

        endingLinkPoint.hookPoint(cg);
        labelName.setPosition(cg);        
    }

    public void draw(Graphics2D g) {

        Point startPoint, endPoint;
        float dash[] = {7.0f};
        Stroke oldStroke = g.getStroke();
        Stroke currentStroke = oldStroke;
        Stroke dashed = new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
                BasicStroke.JOIN_MITER, 10.0f, dash, 0.0f);

        /*
         * Faccio un paio di operazioni per impostare lo stile di tratteggio 
         * giusto in base alla linea da disegnare.         
         */
        switch (stile) {
            case STYLE_IMPLEMENTS:
            case STYLE_USES:
                currentStroke = dashed;
                break;
            default:
                break;
        }

        /*
         * Calcolo il nuovo baricentro tra i punti di aggancio. Ricorda che il
         * baricentro serve da punto di incontro tra le linee che formano un
         * multi arco!
         */
        Point cg = new Point();
        int sumx = 0, sumy = 0;
        for (LinkPoint lp : startLinkPoints) {
            sumx += lp.getAbsoluteX();
            sumy += lp.getAbsoluteY();
        }
        sumx += endingLinkPoint.getAbsoluteX();
        sumy += endingLinkPoint.getAbsoluteY();

        // Media delle coordinate = Baricentro!
        cg.x = sumx / (startLinkPoints.size() + 1);
        cg.y = sumy / (startLinkPoints.size() + 1);

        /*
         * Una volta calcolato ciò che si deve calcolare disegno il tutto...
         * Disegno una linea che parte dalla figura (riferendomi al punto 
         * di aggancio) al baricentro         
         */
        for (LinkPoint lp : startLinkPoints) {
            g.setStroke(currentStroke);
            startPoint = new Point(lp.getAbsoluteX(),
                    lp.getAbsoluteY());
            g.drawLine(
                    startPoint.x, startPoint.y, cg.x, cg.y);
            lp.draw(g);
        }
        endPoint = new Point(endingLinkPoint.getAbsoluteX(),
                endingLinkPoint.getAbsoluteY());
        g.drawLine(
                cg.x, cg.y, endPoint.x, endPoint.y);

        g.setStroke(oldStroke);
        

        /* Disegno infine la cuspide basandomi sull'ultimo segmento che 
         * naturalmente parte dal baricentro e arriva al punto di aggancio
         * della figura finale
         */        
        this.drawCuspide(g, cg, endPoint);
        cg.y-=1;
        labelName.setPosition(cg);
        labelName.draw(g);
        endingLinkPoint.draw(g);

    }

    /**
     * Disegna la cuspide (o punta della freccia) a partire da un segmento
     * orientato dato
     * @param g Contesto grafico su cui viene disegnata la cuspide
     * @param p1 punto iniziale del segmento    
     * @param p2 punto finale del segmento e dove sarà la "punta" della cuspide
     */
    private void drawCuspide(Graphics2D g, Point p1, Point p2) {
        if (p1 == null || p2 == null) {
            return;
        }
        double dx = p1.x - p2.x;
        double dy = p1.y - p2.y;
        double longitud = Math.sqrt(dx * dx + dy * dy);
        double longitudArrow = 15;
        double radioDisminution = 0.50; //0.69999999999999996D;
        boolean whiteCuspide = false;

        switch (stile) {
            case STYLE_HAS:
            case STYLE_USES:
                radioDisminution = 0.50;
                break;
            case STYLE_IMPLEMENTS:
            case STYLE_ISA:
                whiteCuspide = true;
                radioDisminution = 0.70;
                break;
            case STYLE_AGGREGATION:
                whiteCuspide = true;
                break;
            case STYLE_SIMPLEASSOCIATION:
                return;
            default:
                break;
        }

        if (longitud <= longitudArrow) // Check di robustezza
        {
            return;
        }

        Point.Double p3 = new Point.Double(p2.x, p2.y);

        p3.x += (longitudArrow / longitud) * dx;
        p3.y += (longitudArrow / longitud) * dy;

        Point.Double p4 = new Point.Double(p3.x, p3.y);

        p4.x += ((longitudArrow * radioDisminution) / longitud) * dy;
        p4.y -= ((longitudArrow * radioDisminution) / longitud) * dx;
        p3.x -= ((longitudArrow * radioDisminution) / longitud) * dy;
        p3.y += ((longitudArrow * radioDisminution) / longitud) * dx;

        Point.Double p5 = new Point.Double(p2.x, p2.y);

        p5.x += ((longitudArrow / longitud) * dx) * 2;
        p5.y += ((longitudArrow / longitud) * dy) * 2;

        int verticesX[] = {
            (int) p4.x, (int) p2.x, (int) p3.x, (int) p5.x
        };

        int verticesY[] = {
            (int) p4.y, (int) p2.y, (int) p3.y, (int) p5.y
        };

        if (stile == STYLE_HAS || stile == STYLE_USES) {
            g.draw(new Line2D.Double(p2.x, p2.y, p4.x, p4.y));
            g.draw(new Line2D.Double(p2.x, p2.y, p3.x, p3.y));
        } else if (stile == STYLE_COMPOSITION || stile == STYLE_AGGREGATION) {
            Color oldColor = g.getColor();
            if (whiteCuspide) {
                g.setColor(new Color(255, 255, 255));
            }
            g.fillPolygon(verticesX, verticesY, 4);

            g.setColor(oldColor);
            g.draw(new Line2D.Double(p2.x, p2.y, p4.x, p4.y));
            g.draw(new Line2D.Double(p2.x, p2.y, p3.x, p3.y));
            g.draw(new Line2D.Double(p3.x, p3.y, p5.x, p5.y));
            g.draw(new Line2D.Double(p4.x, p4.y, p5.x, p5.y));


        } else {
            Color oldColor = g.getColor();
            if (whiteCuspide) {
                g.setColor(new Color(255, 255, 255));
            }
            g.fillPolygon(verticesX, verticesY, 3);

            g.setColor(oldColor);
            g.draw(new Line2D.Double(p2.x, p2.y, p4.x, p4.y));
            g.draw(new Line2D.Double(p2.x, p2.y, p3.x, p3.y));
            g.draw(new Line2D.Double(p3.x, p3.y, p4.x, p4.y));
        }
    }

    /**
     * Dato un punto, restituisce il punto di aggancio che si trova nel raggio
     * specificato dalla costante RAGGIO_PUNTIAGGANCIO. Un possibile uso di
     * questo metodo e' quello di trovare quale punto di aggancio e' stato
     * selezionato tramite click del mouse.
     * @param reference Punto di riferimento
     * @return -1 se il punto non e' stato trovato, <br>
     * indice intero del punto di aggancio che si trova nel raggio altrimenti     
     */
    public int getNearestLinkPoint(Point reference) {
        int i = 0;
        for (LinkPoint p : startLinkPoints) {
            if (p.getPoint().distance(reference) 
                    <= LINKPOINT_RADIUS) {
                return i;
            }
            i++;
        }
        if (endingLinkPoint.getPoint().distance(reference) 
                <= LINKPOINT_RADIUS) {
            return i;
        }
        return -1;
    }

    /**
     * Dato l'indice intero del punto di aggancio che si vuole spostare e
     * e coordinate x,y qualsiasi trova automaticamente l'aggancio del punto
     * alla relativa figura. Possibile impiego del metodo: al click del mouse
     * su un punto qualsiasi dello spazio di lavoro calcola automaticamente una 
     * posizione valida del punto di aggancio.     
     * @param pointIndex Indice identificativo del punto sul quale agire
     * @param x posizione x qualsiasi del punto di riferimento
     * @param y posizione y qualsiasi del punto di riferimento
     */
    public void setLinkPointPosition(int pointIndex, int x, int y) {
        if (pointIndex > startLinkPoints.size()) {
            return;
        } else if (pointIndex == startLinkPoints.size()) {
            endingLinkPoint.hookPoint(new Point(x, y));
        } else {
            startLinkPoints.get(pointIndex).hookPoint(
                    new Point(x, y));
        }
    }

    /**
     * Imposta la cardinalita' del punto di aggancio identificato da un indice
     * intero positivo.
     * @param pointIndex Indice identificativo del punto sul quale agire
     * @param multiplicity cardinalita' sottoforma di stringa del punto di
     * aggancio
     */
    public void setMultiplicity(int pointIndex, String multiplicity) {
        if (pointIndex > startLinkPoints.size()) {
            return;
        } else if (pointIndex == startLinkPoints.size()) {
            endingLinkPoint.setMultiplicity(multiplicity);
        } else {
            startLinkPoints.get(pointIndex).setMultiplicity(multiplicity);
        }
    }
    
    public void setLabelName(String name)
    {
        labelName.setText(name);
    }
    public int compareTo(Object o) {
        if (this != o) {
            return 1;
        } else {
            return 0;
        }
    }

    /**
     * Metodo che restituisce il numero dei punti di aggancio dell'arco
     * multiplo
     * @return Numero di punti di aggancio dell'arco multiplo
     */
    public int getLinkPointCount() {
        return startLinkPoints.size() + 1;
    }

    /**
     * Ritorna lo stile grafico dell'arco
     * @return stile stile grafico dell'arco
     */
    public int getStyle() {
        return stile;
    }

    /**
     * Imposta lo stile grafico dell'arco
     * @param stile stile grafico da impostare
     */
    public void setStyle(int stile) {
        this.stile = stile;
    }
    
    
    /**
     * Restituisce l'etichetta che rappresenta cardinalita' dell'associazione 
     * identificata da un indice intero dato. E' possibile ottenere tale 
     * etichetta anche per operazioni di scrittura su di essa.
     * @param indice Indice della cardinalita' di interesse
     * @return Cardinalita' dell'associazione identificata dall'intero dato
     */
    /*public Etichetta getEtichettaCardinalita(int indice) {
        if (indice > puntiAggancioIniziali.size()) {
            return null;
        } else if (indice == puntiAggancioIniziali.size()) {
            return puntoAggancioFinale.getEtichettaCardinalita();
        } else {
            return puntiAggancioIniziali.get(indice).getEtichettaCardinalita();
        }
    }*/
}
