/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package it.uniroma1.dis.mfis.redia.verifint.view.graphics.engine;

import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Line2D;

/**
 * Rappresenta un qualsiasi oggetto grafico visualizzabile a schermo che abbia
 * un contorno rettangolare.
 * @author Fuego
 */
public class Figure implements Drawable, Comparable{
    protected Point position;
    protected Dimension dimension;

    /**
     * Crea un'istanza di Figura con posizione in 0,0 e dimensioni 0,0
     */
    public Figure() {
        this.position = new Point(0, 0);
        this.dimension = new Dimension(0, 0);
    }
    
    /**
     * Crea un'istanza di Figura con posizione specificata e dimensioni 0,0
     * @param posizione Posizione della figura
     */
    public Figure(Point posizione) {
        this.position = posizione;
        this.dimension = new Dimension(0, 0);
    }
    
    /**
     * Restituisce la posizione attuale della figura
     * @return posizione attuale della figura
     */
    public Point getPosition() {
        return (Point) position.clone();
    }

    /**
     * Imposta la posizione della figura
     * @param posizione Nuova posizione della figura
     */
    public void setPosition(Point posizione) {
        this.position = posizione;
    }

    
    /**
     * Restituisce le dimensioni attuali della figura
     * @return dimensioni attuali della figura
     */
    public Dimension getDimension() {
        return (Dimension) dimension.clone();
    }

    /**
     * Imposta nuove dimensioni alla figura
     * @param dimensione nuove dimensioni della figura
     */
    public void setDimension(Dimension dimensione) {
        this.dimension = dimensione;
    }
    
    /**
     * Restituisce il punto centrale della figura
     * @return baricentro della figura
     */
    public Point getCG() {
        return new Point((dimension.width / 2) + 
                position.x, (dimension.height / 2) + position.y);
    }   
    
    /**
     * Restituisce un rettangolo in formato x, y, width, height che contiene la
     * figura
     * @return rettangolo che contiene la figura
     */
    public Rectangle getBoundingRect() {
        return new Rectangle(position, dimension);        
    }
    
    /**
     * Restituisce un rettangolo semplice in formato top, left, right, bottom
     * che contiene la figura
     * @return rettangolo semplice che contiene la figura
     */
    public SimpleRect getBoundingSimpleRect() {
        return new SimpleRect(position.x, position.y, position.x 
                + dimension.width, position.y + dimension.height);
    }
    
    public void draw(Graphics2D g) {
        g.drawRect(position.x, position.y, dimension.width, 
                dimension.height);
    }
    
    /**
     * Calcola il determinante di una matrice 2x2
     * @param a elemento 0,0 della matrice
     * @param b elemento 1,0 della matrice
     * @param c elemento 0,1 della matrice
     * @param d elemento 1,1 della matrice
     * @return determinante della matrice
     */
    private double det(double a, double b, double c, double d)
    {
        // Calcolo del determinante su matrice 2x2
        return a * d - b * c;
    }
    
    /**
     * Calcola il punto di intersezione tra due linee (metodo di utilita')
     * @param l1 Prima linea
     * @param l2 Seconda linea
     * @return Intersezione tra le due linee date
     */
    private Point lineIntersection(Line2D l1, Line2D l2){
        Point intersection = new Point();
        if (!l1.intersectsLine(l2))
            return null;

        double  x1 = l1.getX1(), y1 = l1.getY1(),
                x2 = l1.getX2(), y2 = l1.getY2(),
                x3 = l2.getX1(), y3 = l2.getY1(),
                x4 = l2.getX2(), y4 = l2.getY2();

        intersection.x = (int) (det(det(x1, y1, x2, y2), x1 - x2, 
                                det(x3, y3, x4, y4), x3 - x4) / 
                                det(x1 - x2, y1 - y2, x3 - x4, 
                                y3 - y4));
        intersection.y = (int) (det(det(x1, y1, x2, y2), y1 - y2, 
                                det(x3, y3, x4, y4), y3 - y4) / 
                                det(x1 - x2, y1 - y2, x3 - x4, y3 - y4));
        
        return intersection;
    }

    
    /**
     * Calcola il punto di intersezione tra il bordo della figura e una linea
     * data
     * @param linea linea che si interseca con il bordo della figura
     * @return punto di intesezione tra la linea data e il bordo della figura
     */
    public Point getLineIntersection(Line2D linea) {
        /* Attenzione: Si tratta di una funzione ad-hoc che trova soltanto la 
         * prima intersezione, indicato solo per rette che intersecano 
         * un rettangolo  solo una volta!
         * */
        Rectangle rectangle = this.getBoundingRect();
        
        Point intersection = new Point();
        Line2D line1 = new Line2D.Float(new Point(rectangle.x, rectangle.y),
                        new Point(rectangle.x + rectangle.width, 
                        rectangle.y));
        Line2D line2 = new Line2D.Float(new Point(rectangle.x, rectangle.y),
                        new Point(rectangle.x, 
                        rectangle.y  + rectangle.height));
        Line2D line3 = new Line2D.Float(new Point(rectangle.x, rectangle.y + 
                        rectangle.height),new Point(rectangle.x + 
                        rectangle.width, rectangle.y + rectangle.height)); 
        Line2D line4 = new Line2D.Float(new Point(rectangle.x + 
                        rectangle.width, rectangle.y + rectangle.height),
                        new Point(rectangle.x + rectangle.width, 
                        rectangle.y)); 
        
        intersection = lineIntersection(linea, line1);
        if(intersection != null) return intersection;
        intersection = lineIntersection(linea, line2);
        if(intersection != null) return intersection;
        intersection = lineIntersection(linea, line3);
        if(intersection != null) return intersection;
        intersection = lineIntersection(linea, line4);
        if(intersection != null) return intersection;
        
        return null;
    }
    
    
    /**
     * Dato un punto, trova il punto appartenente al bordo della figura
     * più vicino
     * @param p Punto dato
     * @return punto appartenente al bordo della figura più vicino al punto dato
     * @TODO: restituire un punto più credibile nel caso l'intersezione non 
     * esista
     */
    public Point hookPoint(Point p)
    {
        Line2D l = new Line2D.Float(p, this.getCG());
        Point intersezione = getLineIntersection(l);
        if(intersezione == null) {
            return new Point(this.position.x, this.position.y);
        } else return intersezione;
    }
    
    public int compareTo(Object o) {
        if(this != o)
            return 1;
        else return 0;
    }
    
}
